/*
 * StarOffice-Ispell Interface v0.1
 *
 * Integration of `ispell' spell checker into StarOffice.
 *
 * LIBSPELL is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
 * Version 2 (June 1991). See the "COPYING" file distributed with this
 * software for more info.
 *
 * Copyright (c) 2000 by Michael Schroeder <mdhs@gmx.net>
 */
 
#define MAIN

#include "config.h"
#include "ispell.h"
#include "proto.h"
#include "xlang.h"
#include "soling.h"

#include <ctype.h>
#include <fcntl.h>

#ifdef __WINDOWS__
# include <io.h>
# include <stdio.h>
#else
# include <unistd.h>
#endif

#include <sys/stat.h>

#include "version.h"

#if 0
extern struct dent *pershtab;
extern int hcount, pershsize;
#endif

static void initckch (void);

int linit (void);

char CALLTYPE SetLanguage(short nLanguage);
short CALLTYPE Get_Version(void);
char CALLTYPE Available(void);
short CALLTYPE FreeAlloc(void);
char CALLTYPE AddWord(char *Word);
char CALLTYPE SpellQuick(char *Word);
char CALLTYPE SpellWord(char *Word, char **Suggestions);
char CALLTYPE OptionDlg(void *pParent);

static ichar_t token[INPUTWORDLEN + MAXAFFIXLEN];
static int inited = 0;
static short language;					/* actual language */
static char *LibDict = NULL;		/* Pointer to name of $(LIBDIR)/dict */
static char sug_list[MAXPOSSIBLE * MAXSTRINGCHARS];	/* Workarea */


/*
 * StarOffice bug: The word presented to the spell checker sometimes
 * contains non-alphabetic chars.
 * So clean it up here and convert it to canonical form.
 *
 * Legal extra chars (language dependend) are from 'ispell.el'  
 */
static void strcleancnv(char *src)
{
  char c;
  char buf[MAXSTRINGCHARS];
  char *ptr = buf;

  switch (language) {

    case LANGUAGE_GERMAN:
    case LANGUAGE_GERMAN_SWISS:

      while ((c = *src++) != 0) {

        if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')) {
          *ptr++ = c;
          continue;
        }

        switch (c) {
//MS          case 'a' ... 'z':		/* this code is gcc extension... */
//MS          case 'A' ... 'Z':
          case '-':			/* keep boundary char */
          case '\'':			/* other legal char */
          case '\304':			/* umlauts (charset latin1) */
          case '\326':
          case '\334':
          case '\337':
          case '\344':
          case '\366':
          case '\374':
            *ptr++ = c;
	    continue;
        }
      }
      break;

    case LANGUAGE_FRENCH:

      while ((c = *src++) != 0) {

        if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')) {
          *ptr++ = c;
          continue;
        }

        switch (c) {
//MS          case 'a' ... 'z':		/* this code is gcc extension... */
//MS          case 'A' ... 'Z':
          case '\'':			/* other legal char */
          case '\300':			/* charset latin1 ! */
          case '\302':
          case '\306':
          case '\307':
//MS          case '\310' ... '\313':
          case '\316':
          case '\317':
          case '\324':
          case '\331':
          case '\333':
          case '\334':
          case '\340':
          case '\342':
          case '\347':
//MS          case '\350' ... '\353':
          case '\356':
          case '\357':
          case '\364':
          case '\371':
          case '\373':
          case '\374':
            *ptr++ = c;
	    continue;
        }
      }
      break;

/* TODO: more languages */

    default:			/* english... */

      while ((c = *src++) != 0) {

        if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')) {
          *ptr++ = c;
          continue;
        }

        switch (c) {
//MS          case 'a' ... 'z':
//MS          case 'A' ... 'Z':
          case '\'':			/* other legal char */
            *ptr++ = c;
	    continue;
        }
     }
  }
  
  *ptr = '\0';

  strtoichar(token, buf, INPUTWORDLEN * sizeof(ichar_t), 0);
}    


short CALLTYPE Get_Version(void)
{
  return 1;			/* return always 1 */
}


char CALLTYPE Available(void)
{
  return SO_spelling;		/* spell checking is available */
}


char CALLTYPE OptionDlg(void *pParent)	/* Dummy, not used */
{
  pParent = NULL;	/* NULL-Pointer */

  return 0;
}

short CALLTYPE FreeAlloc(void)		/* Dummy, not used */
{
  return SO_true;
}


char CALLTYPE AddWord(char *Word)
{
/* There's a bug in SO: New words are added by only this function.
 * You couldn't add new words in one of your SO extra dictionarys
 * if an external spell checker is active! So we have to add it
 * to the ispell word list here!
 *
 * There's another bug: The argument `*Word' points to waste!
 * So get the word from a former call to SpellWord()....
 * (`token' already contains it)
 */
// treeinsert(Word, strlen(Word), 1);	/* Sorry, this don't work */
 
  treeinsert(ichartosstr(token, 1), ICHARTOSSTR_SIZE, 1);

  treeoutput();		/* write new words to private ispell-dicts */

  return SO_true;
}


char CALLTYPE SpellQuick(char *Word)
{
  if (strchr(Word, '@'))	/* Dont't check e-mail address */
    return SO_true;

  strcleancnv(Word);
  
  return (good(token, 0, 0, 0, 0) || compoundgood(token, 0))? SO_true: SO_false;
}


char CALLTYPE SpellWord(char *Word, char **Suggestions)
{
  int i;
  char *ptr = sug_list;

  if (strchr(Word, '@'))	/* Dont't check e-mail address */
    return SO_true;

  strcleancnv(Word);

  if (good(token, 0, 0, 0, 0) || compoundgood(token, 0))
    return SO_true;

  makepossibilities(token);

  if (pcount) {
    for (i = 0; i < MAXPOSSIBLE;  i++) {
      strcpy(ptr, possibilities[i]);
      ptr += 1 + strlen(possibilities[i]);
      if (possibilities[i][0] == 0) 
        break;
    }
  }
  *ptr = '\0';

  *Suggestions = sug_list;

  return SO_false;
}


char CALLTYPE SetLanguage(short nLanguage)	/* initialize the module */
{
  char *	p;
  char *	preftype = "latin1";
  static char	libdictname[sizeof DEFHASH];

  if (language == nLanguage)	/* already initialized? */
    return SO_true;

  language = nLanguage;

  Trynum = 0;

#if 0
  if (pershtab) {
    free(pershtab);			/* delete private hash table */
    pershtab = NULL;
    pershsize = hcount = 0;
  }
#endif

  if ((p = getenv("STAR_DICTIONARY")) == NULL) {	/* only this dict? */

    switch (nLanguage) {

    case LANGUAGE_DUTCH:
      p = "nederlands.hash";
      break;

    case LANGUAGE_ENGLISH:
      p = "english.hash";
      break;

    case LANGUAGE_ENGLISH_UK:
      p = "british.hash";
      break;

    case LANGUAGE_ENGLISH_US:
      p = "american.hash";
      break;

    case LANGUAGE_FRENCH:
      p = "francais.hash";
      break;

    case LANGUAGE_FINNISH:
      p = "suomi.hash";
      break;

    case LANGUAGE_GERMAN:
    case LANGUAGE_GERMAN_SWISS:			/* OptionDlg required? */
      p = "deutsch.hash";
//      compoundflag = COMPOUND_NEVER;
      compoundflag = COMPOUND_CONTROLLED;
//      compoundflag = COMPOUND_ANYTIME;
//        sortit = 0;
//      tryhardflag = 0;			/* don't guess */
      break;

    case LANGUAGE_ITALIAN:
      p = "italian.hash";
      break;

    case LANGUAGE_NORWEGIAN:
    case LANGUAGE_NORWEGIAN_BOKMAL:
    case LANGUAGE_NORWEGIAN_NYNORSK:
      p = "norsk.hash";
      break;

    case LANGUAGE_PORTUGUESE:
      p = "portugues.hash";
      break;

    case LANGUAGE_SPANISH:
      p = "espanol.hash";
      break;

    case LANGUAGE_SWEDISH:
      p = "svenska.hash";
      break;

    default:    
      p = getenv ("DICTIONARY");
    }
  }

  if (p != NULL) {
    if (index (p, DIRSIGN) != NULL)
      (void) strcpy (hashname, p);
    else
      (void) sprintf (hashname, "%s" DIRSEP "%s", LIBDIR, p);
    (void) strcpy (libdictname, p);
    p = rindex (p, '.');
    if (p == NULL  ||  strcmp (p, HASHSUFFIX) != 0)
      (void) strcat (hashname, HASHSUFFIX);
    LibDict = rindex (libdictname, DIRSIGN);
    if (LibDict != NULL)
      LibDict++;
    else
      LibDict = libdictname;
    p = rindex (libdictname, '.');
    if (p != NULL)
      *p = '\0';
  }
  else
    (void) sprintf (hashname, "%s" DIRSEP "%s", LIBDIR, DEFHASH);

  if (linit () < 0)
    return SO_false;

  if (preftype != NULL) {
    prefstringchar =
      findfiletype (preftype, 1, deftflag < 0 ? &deftflag : (int *) NULL);
    if (prefstringchar < 0)
      return SO_false;
  }

  defdupchar = (prefstringchar < 0)? 0: prefstringchar;

  if (compoundflag < 0)
    compoundflag = hashheader.compoundflag;

  if (tryhardflag < 0)
    tryhardflag = hashheader.defhardflag;

  initckch();

  if (LibDict == NULL) {
    strcpy (libdictname, DEFHASH);
    LibDict = libdictname;
    p = rindex (libdictname, '.');
    if (p != NULL  &&  strcmp (p, HASHSUFFIX) == 0)
      *p = '\0';
  }

  treeinit (NULL, LibDict);

  return SO_true;
}


static void initckch (void)
{
  register ichar_t c;

  for (c = 0; c < (ichar_t) (SET_SIZE + hashheader.nstrchars); ++c)
    {
      if (iswordch (c))
	{
	  if (!mylower (c))
	    {
	      Try[Trynum] = c;
	      ++Trynum;
	    }
	}
      else if (isboundarych (c))
	{
	  Try[Trynum] = c;
	  ++Trynum;
	}
    }
}


/*
 * initialize the hash table
 */
int linit (void)
{
  int			hashfd;
  register int	i;
  register struct dent * dp;
  struct flagent *	entry;
  struct flagptr *	ind;
  int			nextchar;
  int			viazero;
  register ichar_t *	cp;

  if (inited)
    return 0;

  if ((hashfd = open (hashname, 0 | MSDOS_BINARY_OPEN)) < 0)
    {
      return (-1);
    }

  hashsize = read (hashfd, (char *) &hashheader, sizeof hashheader);
  if (hashsize < sizeof hashheader)
    {
      return (-1);
    }
  else if (hashheader.magic != MAGIC)
    {
      return (-1);
    }
  else if (hashheader.magic2 != MAGIC)
    {
      return (-1);
    }
  else if (hashheader.compileoptions != COMPILEOPTIONS
	   ||  hashheader.maxstringchars != MAXSTRINGCHARS
	   ||  hashheader.maxstringcharlen != MAXSTRINGCHARLEN)
    {
      return (-1);
    }

  hashtbl =
  (struct dent *)
      malloc ((unsigned) hashheader.tblsize * sizeof (struct dent));
  hashsize = hashheader.tblsize;
  hashstrings = (char *) malloc ((unsigned) hashheader.stringsize);

  numsflags = hashheader.stblsize;
  numpflags = hashheader.ptblsize;
  sflaglist = (struct flagent *)
    malloc ((numsflags + numpflags) * sizeof (struct flagent));

  if (hashtbl == NULL  ||  hashstrings == NULL  ||  sflaglist == NULL)
    {
      return (-1);
    }

  pflaglist = sflaglist + numsflags;

  if (read (hashfd, hashstrings, (unsigned) hashheader.stringsize)
      != hashheader.stringsize
      || read (hashfd, (char *) hashtbl,
 	    (unsigned) hashheader.tblsize * sizeof (struct dent))
      != hashheader.tblsize * sizeof (struct dent))
    {
      return (-1);
    }


  if (read (hashfd, (char *) sflaglist,
	    (unsigned) (numsflags + numpflags) * sizeof (struct flagent))
      != (numsflags + numpflags) * sizeof (struct flagent))
    {
      return (-1);
    }
  (void) close (hashfd);

   for (i = hashsize, dp = hashtbl;  --i >= 0;  dp++)
     {
       if (dp->word == (char *) -1)
         dp->word = NULL;
       else
         dp->word = &hashstrings [ (int)(dp->word) ];
       if (dp->next == (struct dent *) -1)
         dp->next = NULL;
       else
         dp->next = &hashtbl [ (int)(dp->next) ];
     }

  for (i = numsflags + numpflags, entry = sflaglist; --i >= 0; entry++)
    {
      if (entry->stripl)
	entry->strip = (ichar_t *) &hashstrings[(int) entry->strip];
      else
	entry->strip = NULL;
      if (entry->affl)
	entry->affix = (ichar_t *) &hashstrings[(int) entry->affix];
      else
	entry->affix = NULL;
    }
  /*
  ** Warning - 'entry' and 'i' are reset in the body of the loop
  ** below.  Don't try to optimize it by (e.g.) moving the decrement
  ** of i into the loop condition.
  */
  for (i = numsflags, entry = sflaglist;  i > 0;  i--, entry++)
    {
      if (entry->affl == 0)
	{
	  cp = NULL;
	  ind = &sflagindex[0];
	  viazero = 1;
	}
      else
	{
	  cp = entry->affix + entry->affl - 1;
	  ind = &sflagindex[*cp];
	  viazero = 0;
	  while (ind->numents == 0  &&  ind->pu.fp != NULL)
	    {
	      if (cp == entry->affix)
		{
		  ind = &ind->pu.fp[0];
		  viazero = 1;
		}
	      else
		{
		  ind = &ind->pu.fp[*--cp];
		  viazero = 0;
		}
	    }
	}
      if (ind->numents == 0)
	ind->pu.ent = entry;
      ind->numents++;
      /*
      ** If this index entry has more than MAXSEARCH flags in
      ** it, we will split it into subentries to reduce the
      ** searching.  However, the split doesn't make sense in
      ** two cases:  (a) if we are already at the end of the
      ** current affix, or (b) if all the entries in the list
      ** have identical affixes.  Since the list is sorted, (b)
      ** is true if the first and last affixes in the list
      ** are identical.
      */
      if (!viazero  &&  ind->numents >= MAXSEARCH
	  &&  icharcmp (entry->affix, ind->pu.ent->affix) != 0)
	{
	  /* Sneaky trick:  back up and reprocess */
	  entry = ind->pu.ent - 1; /* -1 is for entry++ in loop */
	  i = numsflags - (entry - sflaglist);
	  ind->pu.fp =
	    (struct flagptr *)
	    calloc ((unsigned) (SET_SIZE + hashheader.nstrchars),
		    sizeof (struct flagptr));
	  if (ind->pu.fp == NULL)
	    {
	      return (-1);
	    }
	  ind->numents = 0;
	}
    }
  /*
  ** Warning - 'entry' and 'i' are reset in the body of the loop
  ** below.  Don't try to optimize it by (e.g.) moving the decrement
  ** of i into the loop condition.
  */
  for (i = numpflags, entry = pflaglist;  i > 0;  i--, entry++)
    {
      if (entry->affl == 0)
	{
	  cp = NULL;
	  ind = &pflagindex[0];
	  viazero = 1;
	}
      else
	{
	  cp = entry->affix;
	  ind = &pflagindex[*cp++];
	  viazero = 0;
	  while (ind->numents == 0  &&  ind->pu.fp != NULL)
	    {
	      if (*cp == 0)
		{
		  ind = &ind->pu.fp[0];
		  viazero = 1;
		}
	      else
		{
		  ind = &ind->pu.fp[*cp++];
		  viazero = 0;
		}
	    }
	}
      if (ind->numents == 0)
	ind->pu.ent = entry;
      ind->numents++;
      /*
      ** If this index entry has more than MAXSEARCH flags in
      ** it, we will split it into subentries to reduce the
      ** searching.  However, the split doesn't make sense in
      ** two cases:  (a) if we are already at the end of the
      ** current affix, or (b) if all the entries in the list
      ** have identical affixes.  Since the list is sorted, (b)
      ** is true if the first and last affixes in the list
      ** are identical.
      */
      if (!viazero  &&  ind->numents >= MAXSEARCH
	  &&  icharcmp (entry->affix, ind->pu.ent->affix) != 0)
	{
	  /* Sneaky trick:  back up and reprocess */
	  entry = ind->pu.ent - 1; /* -1 is for entry++ in loop */
	  i = numpflags - (entry - pflaglist);
	  ind->pu.fp =
	    (struct flagptr *) calloc (SET_SIZE + hashheader.nstrchars,
				       sizeof (struct flagptr));
	  if (ind->pu.fp == NULL)
	    {
	      return (-1);
	    }
	  ind->numents = 0;
	}
    }

  if (hashheader.nstrchartype == 0)
    chartypes = NULL;
  else
    {
      chartypes = (struct strchartype *)
	malloc (hashheader.nstrchartype * sizeof (struct strchartype));
      if (chartypes == NULL)
	{
	  return (-1);
	}
      for (i = 0, nextchar = hashheader.strtypestart;
	   i < hashheader.nstrchartype;
	   i++)
	{
	  chartypes[i].name = &hashstrings[nextchar];
	  nextchar += strlen (chartypes[i].name) + 1;
	  chartypes[i].deformatter = &hashstrings[nextchar];
	  nextchar += strlen (chartypes[i].deformatter) + 1;
	  chartypes[i].suffixes = &hashstrings[nextchar];
	  while (hashstrings[nextchar] != '\0')
	    nextchar += strlen (&hashstrings[nextchar]) + 1;
	  nextchar++;
	}
    }

  inited = 1;

  return 0;
}
