
#include "config.h"
#include "ispell.h"
#include "proto.h"


#define HASHSHIFT   5

#ifdef NO_CAPITALIZATION_SUPPORT
#define HASHUPPER(c)	c
#else /* NO_CAPITALIZATION_SUPPORT */
#define HASHUPPER(c)	mytoupper(c)
#endif /* NO_CAPITALIZATION_SUPPORT */


int hash(ichar_t * word, int hashtblsize);
struct dent *lookup(ichar_t * word, int dotree);


/*
 * The following hash algorithm is due to Ian Dall, with slight modifications
 * by Geoff Kuenning to reflect the results of testing with the English
 * dictionaries actually distributed with ispell.
 */
int hash (s, hashtblsize)
     register ichar_t *	s;
     register int	hashtblsize;
{
  register long	h = 0;
  register int	i;

#ifdef ICHAR_IS_CHAR
  for (i = 4;  i--  &&  *s != 0;  )
    h = (h << 8) | HASHUPPER (*s++);
#else /* ICHAR_IS_CHAR */
  for (i = 2;  i--  &&  *s != 0;  )
    h = (h << 16) | HASHUPPER (*s++);
#endif /* ICHAR_IS_CHAR */
  while (*s != 0)
    {
      /*
       * We have to do circular shifts the hard way, since C doesn't
       * have them even though the hardware probably does.  Oh, well.
       */
      h = (h << HASHSHIFT)
	| ((h >> (32 - HASHSHIFT)) & ((1 << HASHSHIFT) - 1));
      h ^= HASHUPPER (*s++);
    }
  return (unsigned long) h % hashtblsize;
}


/* n is length of s */
struct dent * lookup (s, dotree)
     register ichar_t *		s;
     int				dotree;
{
  register struct dent *	dp;
  register char *		s1;
  char			schar[INPUTWORDLEN + MAXAFFIXLEN];

  dp = &hashtbl[hash (s, hashsize)];

  ichartostr (schar, s, sizeof(schar), 1);

  for (  ;  dp != NULL;  dp = dp->next)
    {
      /* quick strcmp, but only for equality */
      s1 = dp->word;
      if (s1  &&  s1[0] == schar[0]  &&  strcmp (s1 + 1, schar + 1) == 0)
	return dp;

      while (dp->flagfield & MOREVARIANTS)	/* Skip variations */
	dp = dp->next;
    }
  if (dotree)
    {
      dp = treelookup (s);
      return dp;
    }
  else
    return NULL;
}
